#include "wheel_detector.h"
#include <ceres/ceres.h>
#include <ceres/cubic_interpolation.h>

struct WheelCostFunction {
    std::vector<double> line_;
    pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_;

    explicit WheelCostFunction(std::vector<double> &line, pcl::PointCloud<pcl::PointXYZ>::Ptr cloud) {
        line_ = line;
        cloud_ = cloud;
    }

    template<typename JetT>
    bool operator()(const JetT *const x, JetT *residual) const {
        ceres::Grid1D<double> grid(line_.data(), 0, line_.size());
        ceres::CubicInterpolator<ceres::Grid1D<double>> cubic(grid);

        JetT delta_x = x[0];
        JetT theta = x[1];
        auto cos_t = ceres::cos(theta);
        auto sin_t = ceres::sin(theta);
        for (int i = 0; i < cloud_->size(); ++i) {
            JetT cx = JetT(cloud_->points[i].x) * cos_t - JetT(cloud_->points[i].y) * sin_t - delta_x + 1.3;
            cubic.Evaluate(cx * 100.,&residual[i]);
            residual[i]=JetT(1.0)-residual[i];
        }
        return true;
    }
};

singleWheelDetector::singleWheelDetector() {
    float width = 0.1;
    float sigma=100.;
    float solution = 0.01;
    for (int i=-130;i<130;++i) {
        float x = i * solution;
        double y = 1.0 / (1.0 + ceres::exp(-sigma * (x+width/2.))) -
                   1.0 / (1.0 + ceres::exp(-sigma * (x-width/2.)));
        line_.emplace_back(y);
//            printf("---Debug %s %d : [%d %f %f]. \n", __func__, __LINE__, i, x, y);
    }
}


bool singleWheelDetector::detect(pcl::PointCloud<pcl::PointXYZ>::Ptr in_cloud_ptr,
                                 pcl::PointCloud<pcl::PointXYZ>::Ptr out_cloud_ptr,
                                 singleWheelDetector::WheelDetectResult &result) {
    double t_vars[2] = {in_cloud_ptr->points[0].x + 0.1, 0};
    // 构建最小二乘问题
    ceres::Problem problem;

    problem.AddResidualBlock(new ceres::AutoDiffCostFunction<WheelCostFunction, ceres::DYNAMIC, 2>(
                    (new WheelCostFunction(line_, in_cloud_ptr)), in_cloud_ptr->size()),
            nullptr, // 核函数，这里不使用，为空
            t_vars   // 待估计参数
    );

//    problem.SetParameterLowerBound(t_vars, 1, (-20) * M_PI / 180.0f);
//    problem.SetParameterUpperBound(t_vars, 1, (20) * M_PI / 180.0f);

    // 配置求解器
    ceres::Solver::Options options;                             // 这里有很多配置项可以填
    options.linear_solver_type = ceres::DENSE_QR;               // 增量方程如何求解
    options.minimizer_progress_to_stdout = false;               // 输出到cout
    options.max_num_iterations = 100;

    ceres::Solver::Summary summary;                             // 优化信息
    ceres::Solve(options, &problem, &summary);                  // 开始优化
    result.theta = t_vars[1] * 180.0f / M_PI;

    LOG(INFO) << result.info();
    float sin_t = std::sin(t_vars[1]);
    float cos_t = std::cos(t_vars[1]);
    for (auto &point: in_cloud_ptr->points) {
        out_cloud_ptr->emplace_back(point.x * cos_t - point.y * sin_t - t_vars[0],
                                    point.x * sin_t + point.y * cos_t,
                                    point.z);
    }

    return false;
}
